import string, mmap
from utils import *
import kjbuckets
from searching import loop_in_C
import dictzip

class EmptyDb:
    def __init__(self, name):
        self.info = "empty database"
        self.name = name
        
    def match(self, strat, word):
        return []

    def define(self, word):
        return []

class FullDb:
    def __init__(self, name):
        self.info = "full database"
        self.name = name
        
    def match(self, strat, word):
        return [ (self.name, "one"), (self.name, "two") ]

    def define(self, word):
        return [ (self.name, "bla bla blablabla") ]



class FileDb:
    """Entries in index are in UTF8, sorted byte-after-byte
    """
    def __init__(self, *args):
        # if only one parameter, automagically use /usr/share/dict
        # else take it as filenames for database file and index file
        if len(args)==1:
            fname = args[0]
            datafn = "/usr/share/dictd/%s.dict" % fname
            indexfn = "/usr/share/dictd/%s.pyindex" % fname
            self.info = fname
            self.name = fname
        else:
            self.info = args[0]
            self.name = args[0]
            datafn = args[1]
            indexfn = args[2]
        
        try:
            self.datafo = open(datafn)
        except IOError:
            self.datafo = dictzip.DictzipFile(datafn+".dz")
        self.indexfo = open(indexfn)
        self.indexfo.seek(0, 2)
        self.indexlen = self.indexfo.tell()
        self.index = mmap.mmap(self.indexfo.fileno(), self.indexlen, prot=mmap.PROT_READ)
        self.positions = []
        self.index.seek(0)
        while 1:
            p = self.index.tell()
            l = self.index.readline()
            if not l:
                break
            self.positions.append(p)

    def transformentry(self, s):
        # transforms read entry into plain text
        # or, in the future, into mime/html/sgml/whatever
        # to be overriden
        return s

    def readentry(self, arg):
        entry, st, ln = arg
        self.datafo.seek(st)
        r = self.transformentry(self.datafo.read(ln))
        return self.name, r

    def findprefix(self, word):
        low, high = 0, len(self.positions)-1
        middle = high
        while 1:
            if low==middle: #ok, either found or not here...
                r = []
                self.index.seek(self.positions[middle])
                #self.index.readline()
                while 1:
                    l = string.rstrip(self.index.readline())
                    if word==l[:len(word)]:
                        entry, st, ln = string.split(l, TAB)
                        st, ln = b64dec(st), b64dec(ln)
                        r.append( (entry,st,ln) )
                    else:
                        break
                return r
            p = self.positions[middle]
            ap = self.index[p:p+len(word)] # we are here...
            if word > ap:
                low = middle
                middle = low+(high-low)/2
            elif word < ap:
                high = middle
                middle = low+(high-low)/2
            else: # found
                while 1:
                    if middle <= 0:
                        break
                    p = self.positions[middle-1]
                    if word == self.index[p:p+len(word)]:
                        middle -= 1
                    else:
                        break
                low = middle


    def gothroughall(self, word, func, n=10):
        # go through all the words in index, and return first n
        # such that func(indexword, word) returns true
        self.index.seek(0)
        r = []
        while 1:
            l = string.rstrip(self.index.readline())
            if not l:
                break
            entry, st, ln = string.split(l, TAB)
            if func(entry, word):
                st, ln = b64dec(st), b64dec(ln)
                r.append( (entry, st, ln) )
            if len(r)>=n:
                return r
        return r
            

    def define(self, word):
        word = word+TAB
        r = self.findprefix(word)
        r = map(self.readentry, r)
        return r

                
    def match(self, strategy, word):
        r = []
        if strategy=='.':
            strategy = 'exact'
        if strategy in ('exact', 'prefix'):
            if strategy == 'exact':
                word = word+TAB
            r = self.findprefix(word)
        else:
            strategy = strategies[strategy][0]
            # mmap pointer, list of index entries, word, strat, max nr
            r1 = loop_in_C(self.index, self.positions, word, strategy, 10)
            for i in r1:
                self.index.seek(self.positions[i])
                l = self.index.readline()
                entry, st, ln = string.split(l, TAB)
                st, ln = b64dec(st), b64dec(ln)
                r.append( (entry, st, ln) )
        for i in range(len(r)):
            r[i] = self.name, r[i][0]
        r = kjbuckets.kjSet(r).items()
        return r

        
    def __del__(self):
        self.datafo.close()
        self.index.close()
        self.indexfo.close()



class FileDbDict(FileDb):
    """Entries in index are in UTF8, sorted byte-after-byte
       Database file is raw dict file (with %h, %d)
    """

    def transformentry(self, s):
        # transforms read entry into plain text
        rs = string.split(s, '\n')
        reply = []
        i = 0
        while 1:
            srs = string.lstrip(rs[i])
            if srs[:2] == '%h':
                reply.append(srs[2:])
                i+=1
            else:
                break
        if string.strip(rs[i])<>'%d':
            return 'wrong entry, please check'
        i+=1
        reply.append("-----") # separating entry and body
        reply.extend(rs[i:])
        return string.join(reply, '\n')

     