#include <string.h>
#include <ctype.h>
#include <fnmatch.h>
#include <regex.h>
#include "Python.h"
#include "mmapmodule.h"

#include "metaphone.h"
#include "levenshtein.h"
#include "soundex.h"

/* #include <unicode/ustring.h> */


unsigned short int exact_strat(s, w)
char *w, *s;
{
    return (unsigned short int) !strcmp(s, w);
};

unsigned short int substring_strat(s, w)
char *w, *s;
{
    return (unsigned short int) (strstr(s, w) != NULL);
};

unsigned short int suffix_strat(s, w)
char *w, *s;
{
    char *p, *pw;
    pw = w + strlen(w);
    for (p = s + strlen(s) - 1; (p >= s) && (pw > w); p--) {
	pw--;
	if (*pw != *p)
	    return 0;
    }
    return (unsigned short int) (pw <= w);
};


#define SOUNDEX_LEN 8

char *pre_soundex_strat(w)
char *w;
{
    static char hash[SOUNDEX_LEN + 1];
    soundex_hash(w, hash, SOUNDEX_LEN);
    return hash;
};

unsigned short int soundex_strat(s, w)
char *w, *s;
{
    char hash[SOUNDEX_LEN + 1];
    soundex_hash(s, hash, SOUNDEX_LEN);
    return (unsigned short int) !strncmp(hash, w, SOUNDEX_LEN);
};

unsigned short int lev_strat(s, w)
char *w, *s;
{
    return (unsigned short int) (fastest_levdist(s, w) <= 1);
};

char *pre_regexp_strat(w)
char *w;
{
    static regex_t rx;
    regcomp(&rx, w, REG_NOSUB);
    return (char *) &rx;
};

unsigned short int regexp_strat(s, rx)
char *rx, *s;
{
    return (unsigned short int) !regexec((regex_t *) rx, s, 0, NULL, 0);
};

/* if fnmatch does not know about FNM_CASEFOLD, we loose
   case insensitivity. So what - it does not work anyway
   for unicode strings */
#ifndef FNM_CASEFOLD
# define FNM_CASEFOLD 0
#endif

unsigned short int fnmatch_strat(s, w)
char *w, *s;
{
    return (unsigned short int) !fnmatch(w, s, FNM_CASEFOLD);
};


char *pre_metaphone_strat(w)
char *w;
{
    static char metaph[MAXMETAPH + 1];
    metaphone(w, metaph, 1);
    return metaph;
};

unsigned short int metaphone_strat(s, m)
char *s, *m;
{
    return (unsigned short int) !metaphone(s, m, 0);
};


static unsigned short int (*strategies[]) () = {
    &exact_strat, &exact_strat, &substring_strat, &suffix_strat,
	&soundex_strat, &lev_strat, &regexp_strat, &fnmatch_strat,
	&metaphone_strat};

static char *(*pre_strategies[]) () = {
    NULL, NULL, NULL, NULL, &pre_soundex_strat, NULL, &pre_regexp_strat,
	NULL, &pre_metaphone_strat};


static PyObject *loop_in_C(self, args)
PyObject *self;
PyObject *args;
{
    PyObject *mmapobject;
    PyObject *positions;
    char *word;

    PyObject *rlist;
    PyObject *val;


    unsigned short int strategy;
    unsigned int maxnr;

    unsigned int indexlen;
    unsigned int i;
    unsigned long pos;
    char string[255];
    char *spos;
    char *str;
    char *newword;
    unsigned int count;

    if (!PyArg_ParseTuple
	(args, "OOshi", &mmapobject, &positions, &word, &strategy, &maxnr))
	return NULL;
    indexlen = PyList_Size(positions);

    rlist = PyList_New(0);
    if (!rlist) {		/* error */
	PyErr_SetString(PyExc_MemoryError, "error allocating list");
        return NULL;
    };
    count = 0;
    if (pre_strategies[strategy] != NULL)
	newword = pre_strategies[strategy] (word);
    else
	newword = word;
    for (i = 0; i < indexlen && count < maxnr; i++) {
	pos = PyInt_AsLong(PyList_GetItem(positions, i));
	str = (void *) (((mmap_object *) mmapobject)->data + pos);
	spos = memccpy(string, str, '\t', sizeof(string));
	/* string is now entry word in index file */
	if (spos) {
	    *(spos - 1) = '\0';
	} else {
	    printf("Warning: Corrupted index file \n");
	};
	if (strategies[strategy] (string, newword)) {
	    val = Py_BuildValue("i", i);
	    PyList_Append(rlist, val);
	    count++;
	}
    };
    return rlist;
}


static PyMethodDef SearchMethods[] = {
    {"loop_in_C", loop_in_C, METH_VARARGS},
    {NULL, NULL}		/* Sentinel */
};


void initsearching()
{
    (void) Py_InitModule("searching", SearchMethods);
}
